//
// Main interface routines of the HERMES library
// Copyright (c) 1998 Christian Nentwich (brn@eleet.mcb.at)
// This source code is licensed under the GNU LGPL
//
// Please refer to the file COPYING.LIB contained in the distribution for
// licensing conditions
//

#include <string.h>
#include <stdio.h>
#include "Hermes.h"
#include "Base.h"
#include "Conv_GC.h"
#include "Conv_MMX.h"
#include "Conv_x86.h"
#include "Clear_GC.h"
#include "Clearx86.h"
#include "ClearMMX.h"
#include "Head_MMX.h"

// Converter structures
static BaseConverter *HermesConverters[3];  // Array of pointers to all conv.
static BaseConverter *HermesConv;           // The converter to use
static int HermesNumConverters=3;
int32 *Hermes_Lookup;                       // Pointer to aligned lookup table
int32 *Hermes_Original_Lookup=0;            // The pointer to the real lookup
                                            // start
// Clear structures
static BaseClear *HermesCleaner[3];         // Array of clear objects
static BaseClear *HermesClr;                // Clear object to use
static int HermesNumClear=3;

// Reference count
int Hermes_users=0;                         // How many routines are using H.
char Hermes_MMX_available;

void HermesInit()
{ int i;
  int cpuid;

  Hermes_users++;

  if(Hermes_users>1)      // We've already been initialised, just another
  return;                 // class or whatever might be using us too.. 
#ifdef DEBUG
  printf("Hermes initialising: %d\n",Hermes_users);
#endif
  // Create lookup table and align on 16 byte boundary

  Hermes_Original_Lookup=(int32 *)malloc(256*4+15);  
  Hermes_Lookup=Hermes_Original_Lookup;

  while((int)Hermes_Lookup&0xf)
  Hermes_Lookup++;

  // Look for MMX if we compiled x86 assembler support in
#ifndef NO_ASSEMBLER
  cpuid=CheckMMX();

  if(cpuid&0x800000) Hermes_MMX_available=1;
  else Hermes_MMX_available=0;

#endif

  // Assign converters in order

  HermesConverters[0]=&Hermes_MMX;
  HermesConverters[1]=&Hermes_x86;
  HermesConverters[2]=&Hermes_Generic_C;

  // Initialise converters
  
  for(i=0;i<HermesNumConverters;i++)
  HermesConverters[i]->init();  


  // Assign Clear objects in order
  
  HermesCleaner[0]=&Hermes_Clear_MMX;
  HermesCleaner[1]=&Hermes_Clear_x86;
  HermesCleaner[2]=&Hermes_Clear_GC;


  // Initialise clear objects
  for(i=0;i<HermesNumClear;i++)
  HermesCleaner[i]->init();
}


void HermesClose()
{ int i;

  if(Hermes_users) 
  { Hermes_users--;                        // Give up one caller

    if(!Hermes_users)                      // Nobody's using us anymore,destroy
    {
#ifdef DEBUG
      printf("Hermes shutting down\n");
#endif
      for(i=0;i<HermesNumConverters;i++)
      HermesConverters[i]->close();

      for(i=0;i<HermesNumClear;i++)
      HermesCleaner[i]->close();

      if(Hermes_Original_Lookup)
      { free(Hermes_Original_Lookup);
        Hermes_Lookup=0;
      }
    }
  }
}


char HermesRequest(HermesFormat* source, HermesFormat* dest, char stretch)
{ int i;

  HermesConv=0;

  for(i=0;i<HermesNumConverters;i++)
  { if(HermesConverters[i]->request(source,dest,stretch))
    { HermesConv=HermesConverters[i];
      break;
    }
  }

  if(HermesConv==0) return 0;
  else return 1;
}



void HermesPalette(int32 *palette)
{ if(HermesConv)
  HermesConv->palette(palette);
}


void HermesCopy(void *s_pixels,int s_x,int s_y,int s_width,int s_height,
	       int s_pitch,void *d_pixels,int d_x,int d_y,int d_width,
	       int d_height,int d_pitch)
{ if(HermesConv)
  HermesConv->copy(s_pixels,s_x,s_y,s_width,s_height,s_pitch,
	 	   d_pixels,d_x,d_y,d_width,d_height,d_pitch);
}



char HermesRequestClear(HermesFormat* format)
{ int i;

  HermesClr=0;

  // Try all the clear objects in order
  for(i=0;i<HermesNumClear;i++)
  { if(HermesCleaner[i]->request(format))
    { HermesClr=HermesCleaner[i];
      break;
    }
  }

  if(HermesClr==0) return 0;
  else return 1;
}


void HermesClear(void *pixels,int x1,int y1,int width,int height,int pitch,
  int32 r,int32 g,int32 b,char8 index)
{ if(HermesClr)
  HermesClr->clear(pixels,x1,y1,width,height,pitch,r,g,b,index);
}


// Utility routines

void HermesFormatCopy(HermesFormat* source,HermesFormat *dest)
{ memcpy(dest,source,sizeof(HermesFormat));
}


char HermesFormatEquals(HermesFormat* source,HermesFormat *dest)
{ 
  if(source->bpp != dest->bpp) return 0;
  if(source->mask_r != dest->mask_r) return 0;
  if(source->mask_g != dest->mask_g) return 0;
  if(source->mask_b != dest->mask_b) return 0;
  if(source->indexed != dest->indexed) return 0;

  return 1;
}





// utility functions

void Hermes_Calculate_Generic_Info(int32 s_r,int32 s_g,int32 s_b,
  HermesFormat *format,Hermes_Generic_Info *info)
{ int r_right,g_right,b_right;
  
  // Calculate right shift amount for red. If it's <0, then set it to null
  // and calculate left shift amount
  r_right=s_r-Hermes_Topbit(format->mask_r);
  if(r_right<0)
  { info->r_left=-r_right;
    info->r_right=0;
  }
  else
  { info->r_left=0;
    info->r_right=r_right;
  }
   
  g_right=s_g-Hermes_Topbit(format->mask_g);
  if(g_right<0)
  { info->g_left=-g_right;
    info->g_right=0;
  }
  else
  { info->g_left=0;
    info->g_right=g_right;
  }

  b_right=s_b-Hermes_Topbit(format->mask_b);
  if(b_right<0)
  { info->b_left=-b_right;
    info->b_right=0;
  }
  else
  { info->b_left=0;
    info->b_right=b_right;
  }
}


int Hermes_Topbit(int32 mask)
{ int i;

  if(!mask) return 0;

  i=0;

  while(!(mask&1))
  { mask>>=1;
    i++;
  }

  while(mask&1)
  { mask>>=1;
    i++;
  }

  return i;
}
