//
// Window class for PTC 2.0 C++ API
// Copyright (c) 1998 Glenn Fiedler (ptc@gaffer.org)
// This source code is licensed under the GNU LGPL
//

// include files
#include <memory.h>
#include "Core/Error.h"
#include "Win32/Window.h"

// using directive
using namespace ptc::Win32;

// using declarations
using ptc::Core::Error;




Window::Window(HWND window)
{
    // defaults
    defaults();

    // setup window
    m_window = window;
}


Window::Window(const char title[],unsigned extra,unsigned style,int show,int x,int y,int width,int height,void *data)
{
    // defaults
    defaults();

    try
    {
        // save foreground window
        m_foreground = GetForegroundWindow();

        // get module instance
        HINSTANCE instance = GetModuleHandle("ptc.dll");
        if (!instance) instance = GetModuleHandle(0);           

        // create null cursor
        unsigned and_map = 0xFFFFFFFF;
        unsigned xor_map = 0;
        m_cursor = CreateCursor(instance,0,0,1,1,&and_map,&xor_map);
 
        // register window class
        WNDCLASSEX wc;
        wc.cbSize        = sizeof(WNDCLASSEX);
        wc.hInstance     = instance;
        wc.lpszClassName = "PTC_WINDOW";
        wc.lpfnWndProc   = (WNDPROC)WndProc;
        wc.style         = CS_VREDRAW|CS_HREDRAW;
        wc.hIcon         = LoadIcon(instance,"IDI_PTC_ICON");
        wc.hIconSm       = 0;
        wc.hCursor       = m_cursor;//LoadCursor(0,IDC_ARROW);
        wc.lpszMenuName  = 0;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = 0;
        wc.hbrBackground = 0;//GetStockObject(BLACK_BRUSH);
        RegisterClassEx(&wc);

        // setup window data
        strcpy(this->name,"PTC_WINDOW");
        strcpy(this->title,title);
        this->extra  = extra;
        this->style  = style;
        this->show   = show;
        this->x      = x;
        this->y      = y;
        this->width  = width;
        this->height = height;
        this->data   = data;

        // create window event
        m_event = CreateEvent(0,0,0,0);

        // create window thread
        m_thread = CreateThread(0,0,(LPTHREAD_START_ROUTINE)ThreadFunction,this,0,&m_id);

        // wait for window signal
        WaitForSingleObject(m_event,INFINITE);

        // check window handle
        if (!IsWindow(m_window)) throw Error("could not create window");
    }
    catch (Error error)
    {
        // close
        close();

        // error message
        error.rethrow("could not create window");
    }
}


Window::~Window()
{
    // close
    close();
}




HWND Window::handle()
{
    // get handle
    return m_window;
}


DWORD Window::thread()
{
    // get thread id
    return m_id;
}




void Window::defaults()
{
    // defaults
    m_window = 0;
    m_foreground = 0;
    m_cursor = 0;
    m_event = 0;
    m_thread = 0;
    m_id = 0;
}


void Window::close()
{
    // check window and thread
    if (m_thread && IsWindow(m_window))
    {
        // close window
        PostMessage(m_window,WM_DESTROY,0,0);

        // wait for window thread to exit
        WaitForSingleObject(m_thread,INFINITE);
    }

    // restore old foreground window
    if (m_foreground) SetForegroundWindow(m_foreground);

    // release objects
    if (m_event)  CloseHandle(m_event);
    if (m_thread) CloseHandle(m_thread);

    // destroy cursor
    if (m_cursor) DestroyCursor(m_cursor);

    // set normal priority class
    SetPriorityClass(GetCurrentProcess(),NORMAL_PRIORITY_CLASS);
}




void WINAPI Window::ThreadFunction(Window *owner)
{
    // create window
    owner->m_window = CreateWindowEx(owner->extra,owner->name,owner->title,owner->style,owner->x,owner->y,owner->width,owner->height,0,0,0,owner->data);
    
    // check window handle
    if (IsWindow(owner->m_window))
    {
        // show window
        ShowWindow(owner->m_window,owner->show);

        // set window focus
        SetFocus(owner->m_window);

        // set foreground window
        SetForegroundWindow(owner->m_window);

        // signal window event
        SetEvent(owner->m_event);

        // message loop
        MSG message;
        while (GetMessage(&message,0,0,0)==TRUE)
        {
            // process window message
            TranslateMessage(&message);
            DispatchMessage(&message);
        }
    }
    else
    {
        // window failure
        SetEvent(owner->m_event);
    }
}


LRESULT CALLBACK Window::WndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
    // result data
    int result = 0;

    // handle message
    switch (message)
    {
        case WM_DESTROY:
        {
            // shutdown window
            PostQuitMessage(0);
        }
        break;

        case WM_CLOSE:
        {
            // exit
            exit(0);
        }
    
        default:
        {
            // unhandled messages
            result = DefWindowProc(hWnd,message,wParam,lParam);
        }
    }

    // return value
    return result;
}
